#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "chunkhandlers.h"
#include "endian.h"

/* tags */
#define ID_LWO2		'LWO2'
#define	ID_TXUV		'TXUV'

void ReadString(char* out, ChunkContext* ctx) {
	const char* ptr = *ctx->inbuf;
	if(ctx->bytes < ((1+strlen(ptr)+1) & 0x1)) {
		abort();
	}

	uint16_t i=0;
	do {	
		out[i] = ptr[i];
		i++;
	} while(ptr[i-1] != '\0');

	if(i&0x1)	i++;		/* take padding byte into account */
	*(char**)ctx->inbuf += i;
	ctx->bytes -= i;
}

void ReadU16(uint16_t* out, ChunkContext* ctx) {
	if(ctx->bytes < sizeof(uint16_t)) {
		abort();
	}

	const uint16_t* ptr = *ctx->inbuf;
	uint16_t a = *ptr;
	(*(uint16_t**)ctx->inbuf)++;
	ctx->bytes -= sizeof(a);
	#ifdef IS_LITTLEENDIAN
	a =  a<<8 | (a>>8 & 0xFF); 
	#endif
	*out = a;
}

void ReadU32(uint32_t* out, ChunkContext* ctx) {
	if(ctx->bytes < sizeof(uint32_t)) {
		abort();
	}

	const uint32_t* ptr = *ctx->inbuf;
	uint32_t a = *ptr;
	(*(uint32_t**)ctx->inbuf)++;
	ctx->bytes -= sizeof(a);
	#ifdef IS_LITTLEENDIAN
	a =  a<<24 | (a<<8 & 0x00FF0000) | (a>>8 & 0xFF00) | (a>>24 & 0xFF); 
	#endif
	*out = a;
}

void ReadFloat4(float* out, ChunkContext* ctx) {
	ReadU32((uint32_t*)out, ctx);
}

void HandleDefault(ChunkContext* ctx) {
	if(ctx->size > ctx->bytes)
		abort();
	*(uint8_t**)ctx->inbuf += ctx->size;
	ctx->bytes -= ctx->size;
}

void HandleFORM(ChunkContext* ctx) {
	uint32_t tag;
	ReadU32(&tag, ctx);
	if(ID_LWO2 != tag) {
		fprintf (stderr, "Not a LWO2 file.\n" );
		exit(1);
	}
	ChunkContext new_context = {
		.parser = ctx->parser,
		.inbuf = ctx->inbuf,
		.bytes = ctx->size-sizeof(uint32_t),
		.parentbytes = &ctx->bytes,
		.issubchunk = false,
		.data = NULL,
		.layer_number = ctx->layer_number };
	if(false == ReadChunks(&new_context))
		abort();
}

void HandleSURF(ChunkContext* ctx) {
	size_t len1 = (strlen(*(const char**)ctx->inbuf)+2) & 0xFFFFFFFE;	// offset to the 'source' string
	size_t len2 = strlen(*(const char**)ctx->inbuf + len1) + 2;

	char* strname;
	char* strsource;
	if(NULL == (strname = malloc(len1)) || NULL == (strsource = malloc(len2))) {
		abort();
	}

	uint32_t tmp = ctx->bytes;
	ReadString(strname, ctx);
	ReadString(strsource, ctx);

	Surface* surface = surfaceNew(strname, strsource, NULL);

	ChunkContext new_context = {
		.parser = ctx->parser,
		.inbuf = ctx->inbuf,
		.bytes = ctx->size - (tmp-ctx->bytes),
		.parentbytes = &ctx->bytes,
		.issubchunk = true,
		.data = surface,
		.layer_number = ctx->layer_number
	};
	if(!ReadChunks(&new_context))
		return;
	meshAddSurface(&ctx->parser->mesh, surface);
}

void HandleTAGS(ChunkContext* ctx) {
	char str[ctx->size];
	while(ctx->bytes > ctx->bytesafter) {
		ReadString(str, ctx);
		char* tag = malloc(strlen(str)+1);
		if(NULL == tag) {
			abort();
		}
		strcpy(tag, str);
		vecPushBack(Pchar, &ctx->parser->mesh.tags, tag);
	}
}

void HandlePNTS(ChunkContext* ctx) {
	float x,y,z;
	Vertex vtx;
	for(uint32_t i=0, n=ctx->size/12; i<n; i++) {
		ReadFloat4(&x, ctx);
		ReadFloat4(&y, ctx);
		ReadFloat4(&z, ctx);

		vector3fSet(&vtx.pos, x, y, z);
		layerAddVertex(ctx->layer, &vtx);
	}
}

void HandlePOLS(ChunkContext* ctx) {
	uint32_t type;
	ReadU32(&type, ctx);
				
	uint16_t nvts;
	Face face;
	while(ctx->bytes > ctx->bytesafter) {
		ReadU16(&nvts, ctx);
		if((nvts &= 0x3FF) > FACE_MAX_VTS)
			abort();

		faceConstruct(&face, nvts);

		uint16_t vi16;
		for(uint16_t i=0; i<nvts; i++) {
			ReadU16(&vi16, ctx);
			if(vi16 < 0xFF00) {
				faceSetIndex(&face, i, vi16);
			} else {
				fprintf (stderr, "error: currently there's support for no more that 0xFF00 verts\n" );
				exit(1);
			}
		}
		layerAddFace(ctx->layer, &face);
	}
}

void HandlePTAG(ChunkContext* ctx) {
	uint32_t type;
	ReadU32(&type, ctx);

	uint16_t polyidx;
	uint16_t polytag;
	while(ctx->bytes > ctx->bytesafter) {
		ReadU16(&polyidx, ctx);
		if(polyidx >= 0xFF00)
			abort();		/* currently only 'small' indexes */
		ReadU16(&polytag, ctx);

		Pair ptpair;
		pairConstruct(&ptpair, polyidx, polytag);
		layerAddPair(ctx->layer, &ptpair);
	}
}

void HandleLAYR(ChunkContext* ctx) {
	uint32_t bytesbefore = ctx->bytes;
	uint16_t number;
	ReadU16(&number, ctx);

	uint16_t flags;
	ReadU16(&flags, ctx);

	Vector3f pivot;
	ReadFloat4(&pivot.x, ctx);
	ReadFloat4(&pivot.y, ctx);
	ReadFloat4(&pivot.z, ctx);

	char name[ctx->size];
	ReadString(name, ctx);

	uint16_t parent = -1;
	if(ctx->size - (bytesbefore - ctx->bytes))
		ReadU16(&parent, ctx);

	Layer* new_layer = layerNew(number, flags, &pivot, name, parent);
	meshAddLayer(&ctx->parser->mesh, new_layer);
	ctx->layer = new_layer;
}

void HandleVMAP(ChunkContext* ctx) {
	uint32_t type;
	ReadU32(&type, ctx);
	if(ID_TXUV != type) {
		printf ( "warning: unsupported VMAP type.\n" );
	}
	uint16_t dim;
	ReadU16(&dim, ctx);
	if(2 != dim) {
		printf ( "warning: unsupported number of texture dimmensions.\n" );
	}
	char name[256];
	ReadString(name, ctx);

	float texcoords[dim];
	uint16_t vi16;
	while(ctx->bytes > ctx->bytesafter) {
		ReadU16(&vi16, ctx);
		if(vi16 >= 0xFF00) {
			abort();
		}
		for(uint16_t i=0; i<dim; i++) {
			ReadFloat4(&texcoords[i], ctx);
		}
		Vertex* vtx = layerGetVertex(ctx->layer, vi16);
		Vector2f* tc = vertexGetTexCoords(vtx);
		vector2fSet(tc, texcoords[0], texcoords[1]);
	}
}

void HandleVMAD(ChunkContext* ctx) {
	uint32_t type;
	ReadU32(&type, ctx);
	if(ID_TXUV != type) {
		printf ( "warning: unsupported VMAP type.\n" );
	}
	uint16_t dim;
	ReadU16(&dim, ctx);
	if(2 != dim) {
		printf ( "warning: unsupported number of texture dimmensions.\n" );
	}
	char name[256];
	ReadString(name, ctx);

	float texcoords[dim];
	uint16_t vi16, pi16;
	while(ctx->bytes > ctx->bytesafter) {
		ReadU16(&vi16, ctx);
		if(vi16 >= 0xFF00) {
			abort();
		}

		ReadU16(&pi16, ctx);
		if(pi16 >= 0xFF00) {
			abort();
		}

		for(uint16_t i=0; i<dim; i++) {
			ReadFloat4(&texcoords[i], ctx);
		}

		Face* face = layerGetFace(ctx->layer, pi16);
		uint16_t i,n;
		for(i=0, n=faceGetVertexCount(face); i<n; i++) {
			if(faceGetIndex(face, i) == vi16)
				break;
		}
		assert(n != i);

		/* duplicate vertex */
		layerAddVertex(ctx->layer, layerGetVertex(ctx->layer, vi16));
		uint32_t lastvi = layerGetVertexCount(ctx->layer)-1;
		Vertex* vtx = layerGetVertex(ctx->layer, lastvi);
		Vector2f* tc = vertexGetTexCoords(vtx);
		vector2fSet(tc, texcoords[0], texcoords[1]);
		/* update face so now it refers to it */
		faceSetIndex(face, i, lastvi);
	}
}

void HandleCLIP(ChunkContext* ctx) {
	uint32_t index;
	ReadU32(&index, ctx);

	Clip* clip = clipNew(index, NULL);

	ChunkContext new_context = {
		.parser = ctx->parser,
		.inbuf = ctx->inbuf,
		.bytes = ctx->size - sizeof(index),
		.parentbytes = &ctx->bytes,
		.issubchunk = true,
		.data = clip,
		.layer_number = ctx->layer_number,
		.layer = ctx->layer,
	};
	if(!ReadChunks(&new_context))
		abort();
	meshAddClip(&ctx->parser->mesh, clip);
}

void HandleCOLR_sub(ChunkContext* ctx) {
	Vector3f color;
	ReadFloat4(&color.x, ctx);
	ReadFloat4(&color.y, ctx);
	ReadFloat4(&color.z, ctx);

	uint16_t envelope;
	ReadU16(&envelope, ctx);
	if(envelope >= 0xFF00)
		abort();		/* currently we support only 'small' indexes */
	surfaceSetColor((Surface*)ctx->data, &color);
}

void HandleSIZE_sub(ChunkContext* ctx) {
	Block* block = (Block*)ctx->data;
	Vector3f vec;
	uint16_t envel16;

	ReadFloat4(&vec.x, ctx);
	ReadFloat4(&vec.y, ctx);
	ReadFloat4(&vec.z, ctx);
	ReadU16(&envel16, ctx);
	if(envel16 >= 0xFF00)
		abort();
	block->tmap.size.vec = vec;
	block->tmap.size.envel = envel16;
}

void HandleBLOK_sub(ChunkContext* ctx) {
	BlockHeader bheader;
	Block* block = blockNew(&bheader);
	surfaceAddBlock((Surface*)ctx->data, block);

	ChunkContext new_context = {
		.parser = ctx->parser,
		.inbuf = ctx->inbuf,
		.bytes = ctx->size,
		.parentbytes = &ctx->bytes,
		.issubchunk = true,
		.data = block,
		.layer_number = ctx->layer_number,
		.layer = ctx->layer,
	};
	if(!ReadChunks(&new_context))
		abort();
}

void HandleIMAP_sub(ChunkContext* ctx) {
	Block* block = (Block*)ctx->data;
	block->head.ordinal = malloc(strlen(*ctx->inbuf)+1);

	ReadString(block->head.ordinal, ctx);

	if(!ReadChunks(ctx))
		abort();
}

void HandleTMAP_sub(ChunkContext* ctx) {
	ChunkContext new_context = {
		.parser = ctx->parser,
		.inbuf = ctx->inbuf,
		.bytes = ctx->size,
		.parentbytes = &ctx->bytes,
		.issubchunk = true,
		.data = ctx->data,
		.layer_number = ctx->layer_number,
		.layer = ctx->layer,
	};
	if(!ReadChunks(&new_context))
		abort();
}

void HandleIMAG_sub(ChunkContext* ctx) {
	Block* block = (Block*)ctx->data;
	uint16_t clip_id_16;
	ReadU16(&clip_id_16, ctx);
	if(clip_id_16 >= 0xFF00)
		abort();
	block->imag.clip_id = clip_id_16;
}

void HandleSTIL_sub(ChunkContext* ctx) {
	char* name = malloc(strlen(*ctx->inbuf) + 1);
	ReadString(name, ctx);
	((Clip*)ctx->data)->name = name;
}

