#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include "converter.h"
#include "hash.h"
#include "file.h"
#include "image.h"
#include "vec_pvoid.h"
#include "utils.h"

extern Vec(Pvoid) g_imagevec;

#include "pairclipimage.h"

void LoadImages(const Mesh* mesh) {
	for(uint32_t i=0, n=meshGetClipCount(mesh); i<n; i++) {
		const Clip* clip = meshGetConstClip(mesh, i);
		char* path = fileExtractTillLast(mesh->filename, '/');
		char* filepath = malloc(strlen(path) + strlen(clip->name) + 1);
		strcpy(filepath, path);
		strcat(filepath, fileExtractAfterLast(clip->name, '/'));
		free(path);

		Image* image = imageNew(filepath);
		if(NULL == image) {
			printf ( "Warning: Couldn't open %s\n", filepath );
		} 
		PairClipImage* pair = malloc(sizeof(PairClipImage));
		pair->clip = clip;
		pair->image = image;
		vecPushBack(Pvoid, &g_imagevec, pair);
	
		free(filepath);
	}
}

void converterProcessFaces(UltraLayer* ulayer, const Mesh* mesh, const Layer* layer) {
	if(mesh) {}
	Hash* hash = layerSpecifyGroups(layer);
	
	uint32_t n = hashGetNumKeys(hash);
	if(n > 1)
		printf ( "warning: More than 1 (%d) PTAG.tag values in layer %d\n", n, layerGetNumber(layer) );
	for(uint32_t i=0; i<n; i++) {
		const MyPair* pair = *vecGetConstElement(PMyPair, &hash->pairs, i);
		const Vec(uint32_t)* polyids = &pair->vals;

		for(uint32_t i=0, pnum=vecGetSize(uint32_t, polyids); i<pnum; i++) {
			uint32_t poly = *vecGetConstElement(uint32_t, polyids, i);
			const Face* face = layerGetConstFace(layer, poly);
			uint16_t nvts = faceGetVertexCount(face);
			if(nvts > FACE_MAX_VTS) {
				fprintf (stderr, "No support for more that %d verts in a face\n", FACE_MAX_VTS );
				exit(1);
			}
		
			UltraFace uface;
			if(3 == nvts)
				ultrafaceSetTriangle(&uface, faceGetIndex(face, 0), faceGetIndex(face, 1), faceGetIndex(face, 2));
			else if(4 == nvts)
				ultrafaceSetQuad(&uface, faceGetIndex(face, 0), faceGetIndex(face, 1), faceGetIndex(face, 2), faceGetIndex(face, 3));
			else {
				fprintf (stderr, "Support only for 3 or 4 vertices in a face\n" );
				exit(1);
			}
			ultralayerAddFace(ulayer, &uface);
		}
	}
	
	/* for(uint32_t i=0, n=hashGetNumKeys(&hash); i<n; i++) {
		const MyPair* pair = *vecGetConstElement(PMyPair, &hash.pairs, i);
		vecchar_sprintf(tempvec, "\t{&materials[%d], drawmesh%d_%d},\n", pair->key, layerGetNumber(layer), i);
	} */

	hashDelete(hash);
}

void converterProcessVerts(UltraLayer* ulayer, const Mesh* mesh, const Layer* layer) {
	for(uint32_t i=0, n=layerGetVertexCount(layer); i<n; i++) {
		const Vertex* vtx = layerGetConstVertex(layer, i);
		UltraVertex uvtx;
	
		const Vector3f* pos = vertexGetConstPosition(vtx);
		if(pos->x < -32767.0 || pos->x > 32767.0 ||
		   pos->y < -32767.0 || pos->y > 32767.0 ||
		   pos->z < -32767.0 || pos->z > 32767.0) {
			printf ( "warning: Vertex %d in layer %d has coordinate(s) not fitting in 16 bits.\n", i, layerGetNumber(layer));
		}
		/* find a polygon containing this vertex */
		uint32_t fidx, m;
		for(fidx=0, m=layerGetFaceCount(layer); fidx<m; fidx++) {
			const Face* face = layerGetConstFace(layer, fidx);
			if(faceRefersTo(face, i))
				break;
			if(m-1 == fidx) {
				abort();
			}
		}
		/* find a surface associated with this polygon */
		uint32_t tidx;
		for(uint32_t pidx=0, m=layerGetPairCount(layer); pidx<m; pidx++) {
			const Pair* pair = layerGetConstPair(layer, pidx);
			tidx = pair->tagidx;
			if(pair->polyidx == fidx)
				break;
			if(m-1 == pidx) {
				abort();
			}
		}
		const Surface* surface = meshGetConstSurfaceByName(mesh, meshGetConstTag(mesh, tidx));
		/* apply texture information to the vertex */
		float stw, sth;
		uint32_t tw, th;
		const Block* block = NULL;
		if(surfaceGetBlockCount(surface)) {
			block = surfaceGetConstBlock(surface, 0);
			stw = block->tmap.size.vec.x;
			sth = block->tmap.size.vec.y;
			const Clip* clip = meshGetConstClipById(mesh, blockGetClipId(block));
			if(clip) {
				for(uint32_t i=0, n=vecGetSize(Pvoid, &g_imagevec); i<n; i++) {
					const PairClipImage* pair = *vecGetConstElement(Pvoid, &g_imagevec, i);
					if(pair->clip == clip) {
						if(pair->image) {
							tw = pair->image->width;
							th = pair->image->height;
							break;
						}
					}
				}
			} else {
				printf ( "Warning: couldn't find Clip %d in Block of Surface %s\n", blockGetClipId(block), meshGetConstTag(mesh, tidx) );
			}
		} 

		ultravertexSetPosition(&uvtx, pos->x, pos->y, pos->z);
	
		const Vector3f* norm = vertexGetConstNormal(vtx);
		ultravertexSetNormal(&uvtx, 127*norm->x, 127*norm->y, 127*norm->z);
	
		const Vector2f* tc = vertexGetConstTexCoords(vtx);
		ultravertexSetTexcoords(&uvtx, stw*2*32*tw*tc->x, sth*2*32*th*tc->y);
 	
		ultralayerAddVertex(ulayer, &uvtx);
	}
}

void converterProcessLayers(UltraMesh* umesh, const Mesh* mesh) {
	for(uint32_t i=0, n=meshGetLayerCount(mesh); i<n; i++) {
		const Layer* layer = meshGetConstLayer(mesh, i);
		const Hash* hash = layerSpecifyGroups(layer);
		const MyPair* pair = *vecGetConstElement(PMyPair, &hash->pairs, 0);

		UltraLayer ulayer;
		ultralayerConstruct(&ulayer, pair->key);
	
		converterProcessVerts(&ulayer, mesh, layer);
		converterProcessFaces(&ulayer, mesh, layer);
	
		ultrameshAddLayer(umesh, &ulayer);
	}
}

void converterProcessClips(UltraMesh* umesh, const Mesh* mesh) {
	for(uint32_t i=0, n=meshGetClipCount(mesh); i<n; i++) {
		const Clip* clip = meshGetConstClip(mesh, i);
		ultrameshAddTexture(umesh, clipGetId(clip));
	}
}

void converterProcessSurfaces(UltraMesh* umesh, const Mesh* mesh) {
	for(uint32_t i=0, n=meshGetSurfaceCount(mesh); i<n; i++) {
		const Surface* surf = meshGetConstSurface(mesh, i);
		UltraMaterial umaterial;

		const Block* block = NULL;
		uint32_t texid = ULTRAMATERIAL_TEXTURE_NONE;
		for(uint32_t i=0, n=surfaceGetBlockCount(surf); i<n; i++) {
			block = surfaceGetConstBlock(surf, i);
		}
		if(block) {
			texid = blockGetClipId(block)-1;
		}

		ultramaterialSet(&umaterial, surfaceGetConstColor(surf), texid);
		ultrameshAddMaterial(umesh, &umaterial);
	}
}

UltraMesh* converterProcess(const Mesh* mesh) {
	UltraMesh* umesh = ultrameshNew();
	
	LoadImages(mesh);
	converterProcessLayers(umesh, mesh);
	converterProcessClips(umesh, mesh);
	converterProcessSurfaces(umesh, mesh);

	return umesh;
}

