#include <stdlib.h>
#include <string.h>
#include "layer.h"
#include "hash.h"
#include "math.h"

inline Layer* layerNew(uint16_t number, uint16_t flags, Vector3f* pivot, char* name, uint16_t parent) {
	Layer* layer = malloc(sizeof(Layer));
	if(NULL == layer)
		abort();
	layerConstruct(layer, number, flags, pivot, name, parent);
	return layer;
}

inline void layerDelete(Layer* layer) {
	layerDestroy(layer);
	free(layer);
}

void layerConstruct(Layer* layer, uint16_t number, uint16_t flags, Vector3f* pivot, char* name, uint16_t parent) {
	layer->number = number;
	layer->flags = flags;
	layer->pivot = *pivot;
	layer->name = strcpy(malloc(strlen(name)+1), name);
	layer->parent = parent;

	vecConstruct(Vertex, &layer->vts);
	vecConstruct(Face, &layer->pols);
	vecConstruct(Pair, &layer->ptpairs);
}

void layerDestroy(Layer* layer) {
	free(layer->name);
	vecDestroy(Pair, &layer->ptpairs);
	vecDestroy(Face, &layer->pols);
	vecDestroy(Vertex, &layer->vts);
}

inline uint16_t layerGetNumber(const Layer* layer) {
	return layer->number;
}

inline uint32_t layerGetVertexCount(const Layer* layer) {
	return vecGetSize(Vertex, &layer->vts);
}

inline uint32_t layerGetFaceCount(const Layer* layer) {
	return vecGetSize(Face, &layer->pols);
}

inline uint32_t layerGetPairCount(const Layer* layer) {
	return vecGetSize(Pair, &layer->ptpairs);
}

inline Vertex* layerGetVertex(Layer* layer, uint32_t i) {
	return vecGetElement(Vertex, &layer->vts, i);
}

inline Face* layerGetFace(Layer* layer, uint32_t i) {
	return vecGetElement(Face, &layer->pols, i);
}

inline Pair* layerGetPair(Layer* layer, uint32_t i) {
	return vecGetElement(Pair, &layer->ptpairs, i);
}

inline const Vertex* layerGetConstVertex(const Layer* layer, uint32_t i) {
	return vecGetConstElement(Vertex, &layer->vts, i);
}

inline const Face* layerGetConstFace(const Layer* layer, uint32_t i) {
	return vecGetConstElement(Face, &layer->pols, i);
}

inline const Pair* layerGetConstPair(const Layer* layer, uint32_t i) {
	return vecGetConstElement(Pair, &layer->ptpairs, i);
}

inline void layerAddVertex(Layer* layer, Vertex* vertex) {
	vecPushBack(Vertex, &layer->vts, *vertex);
}

inline void layerAddFace(Layer* layer, Face* face) {
	vecPushBack(Face, &layer->pols, *face);
}

inline void layerAddPair(Layer* layer, Pair* pair) {
	vecPushBack(Pair, &layer->ptpairs, *pair);
}

void layerCalculateNormals(Layer* layer) {
	Hash hash;
	hashConstruct(&hash);

	for(uint32_t i=0, n=layerGetFaceCount(layer); i<n;  i++) {
		const Face* face = layerGetConstFace(layer, i);
		for(uint16_t j=0, m=faceGetVertexCount(face); j<m; j++) {
			hashInsert(&hash, faceGetIndex(face, j), i);
		}
	}

	for(uint32_t i=0; i<hashGetNumKeys(&hash); i++) {
		MyPair* pair = hashGetPair(&hash, i);
		Vector3f vnormal = { 0, 0, 0 };
		for(uint32_t j=0, n=vecGetSize(uint32_t, &pair->vals); j<n; j++) {
			const Face* face = layerGetConstFace(layer, *vecGetElement(uint32_t, &pair->vals, j));
			const Vertex* p0 = layerGetConstVertex(layer, faceGetIndex(face, 0));
			const Vertex* p1 = layerGetConstVertex(layer, faceGetIndex(face, 1));
			const Vertex* p2 = layerGetConstVertex(layer, faceGetIndex(face, 2));

			Vector3f u,v;
			vector3fSub(&u, vertexGetConstPosition(p1), vertexGetConstPosition(p0));
			vector3fSub(&v, vertexGetConstPosition(p2), vertexGetConstPosition(p0));
			Vector3f fnormal;
			mathCrossProduct(&fnormal, &u, &v);
			vector3fAdd(&vnormal, &vnormal, &fnormal);
		}
		mathNormalizeVector(&vnormal, &vnormal);

		Vertex* vtx = layerGetVertex(layer, pair->key);
		*vertexGetNormal(vtx) = vnormal;
	}
	hashDestroy(&hash);
}

Hash* layerSpecifyGroups(const Layer* layer) {
	Hash* hash = hashNew();

	/* group faces by shared surface tags */
	for(uint32_t i=0, n=layerGetPairCount(layer); i<n; i++) {
		const Pair* pair = layerGetConstPair(layer, i);
		hashInsert(hash, pair->tagidx, pair->polyidx);
	}
	
	return hash;
}

