#include <string.h>
#include <stdbool.h>
#include "mesh.h"

void meshConstruct(Mesh* mesh) {
	vecConstruct(Pchar, &mesh->tags);
	vecConstruct(PLayer, &mesh->layers);
	vecConstruct(PSurface, &mesh->surfs);
	vecConstruct(PClip, &mesh->clips);
}

void meshDestroy(Mesh* mesh) {
	for(uint32_t i=0; i<meshGetClipCount(mesh); i++) {
		clipDelete(*vecGetElement(PClip, &mesh->clips, i));
	}
	for(uint32_t i=0; i<meshGetSurfaceCount(mesh); i++) {
		surfaceDelete(*vecGetElement(PSurface, &mesh->surfs, i));
	}
	for(uint32_t i=0; i<meshGetLayerCount(mesh); i++) {
		layerDelete(*vecGetElement(PLayer, &mesh->layers, i));
	}
	for(uint32_t i=0; i<vecGetSize(Pchar, &mesh->tags); i++) {
		free(*vecGetElement(Pchar, &mesh->tags, i));
	}
	vecDestroy(PClip, &mesh->clips);
	vecDestroy(PSurface, &mesh->surfs);
	vecDestroy(PLayer, &mesh->layers);
	vecDestroy(Pchar, &mesh->tags);
}

void meshScale(Mesh* mesh, double scale) {
	for(uint16_t i=0, n=meshGetLayerCount(mesh); i<n; i++) {
		Layer* layer = meshGetLayer(mesh, i);
		for(uint32_t j=0, n=layerGetVertexCount(layer); j<n; j++) {
			Vertex* vtx = layerGetVertex(layer, j);
			Vector3f* pos = vertexGetPosition(vtx);
			vector3fScaleUniform(pos, scale);
		}
	}
}

inline uint16_t meshGetLayerCount(const Mesh* mesh) {
	return vecGetSize(PLayer, &mesh->layers);
}

inline uint32_t meshGetVertexCount(const Mesh* mesh) {
	uint32_t n=0;
	for(uint32_t i=0; i<meshGetLayerCount(mesh); i++) {
		const Layer* layer = *vecGetConstElement(PLayer, &mesh->layers, i);
		n += layerGetVertexCount(layer);
	}
	return n;
}

inline uint32_t meshGetFaceCount(const Mesh* mesh) {
	uint32_t n=0;
	for(uint32_t i=0; i<meshGetLayerCount(mesh); i++) {
		const Layer* layer = *vecGetConstElement(PLayer, &mesh->layers, i);
		n += layerGetFaceCount(layer);
	}
	return n;
}

inline uint32_t meshGetSurfaceCount(const Mesh* mesh) {
	return vecGetSize(PSurface, &mesh->surfs);
}

inline uint32_t meshGetClipCount(const Mesh* mesh) {
	return vecGetSize(PClip, &mesh->clips);
}

Layer* meshGetLayer(Mesh* mesh, uint32_t i) {
	return *vecGetElement(PLayer, &mesh->layers, i);
}

Layer* meshGetLayerByNumber(Mesh* mesh, const uint16_t number) {
	for(uint32_t i=0, n=meshGetLayerCount(mesh); i<n; i++) {
		Layer* layer = *vecGetElement(PLayer, &mesh->layers, i);
		if(number == layer->number)
			return layer;
	}

	return NULL;
} 

Surface* meshGetSurface(Mesh* mesh, uint32_t i) {
	return *vecGetElement(PSurface, &mesh->surfs, i);
}

Clip* meshGetClip(Mesh* mesh, uint32_t i) {
	return *vecGetElement(PClip, &mesh->clips, i);
}

const Layer* meshGetConstLayer(const Mesh* mesh, uint32_t i) {
	return *vecGetConstElement(PLayer, &mesh->layers, i);
}

const Surface* meshGetConstSurface(const Mesh* mesh, uint32_t i) {
	return *vecGetConstElement(PSurface, &mesh->surfs, i);
}

const Surface* meshGetConstSurfaceByName(const Mesh* mesh, const char* name) {
	for(uint32_t i=0, n=meshGetSurfaceCount(mesh); i<n; i++) {
		const Surface* surface = meshGetConstSurface(mesh, i);
		if(0 == strcmp(surfaceGetName(surface), name))
			return surface;
	}
	return NULL;
}

const Clip* meshGetConstClip(const Mesh* mesh, uint32_t i) {
	return *vecGetConstElement(PClip, &mesh->clips, i);
}

const Clip* meshGetConstClipById(const Mesh* mesh, uint32_t id) {
	for(uint32_t i=0, n=meshGetClipCount(mesh); i<n; i++) {
		const Clip* clip = meshGetConstClip(mesh, i);
		if(clipGetId(clip) == id)
			return clip;
	}
	return NULL;
}

const char* meshGetConstTag(const Mesh* mesh, uint32_t i) {
	return *vecGetConstElement(Pchar, &mesh->tags, i);
}

void meshAddLayer(Mesh* mesh, Layer* layer) {
	vecPushBack(PLayer, &mesh->layers, layer);
}

void meshAddSurface(Mesh* mesh, Surface* surface) {
	vecPushBack(PSurface, &mesh->surfs, surface);
}

void meshAddClip(Mesh* mesh, Clip* clip) {
	vecPushBack(PClip, &mesh->clips, clip);
}
