#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include "ultramesh.h"
#include "utils.h"

UltraMesh* ultrameshNew(void) {
	UltraMesh* umesh = malloc(sizeof(UltraMesh));
	if(NULL == umesh)
		abort();
	ultrameshConstruct(umesh);
	return umesh;
}

void ultrameshDelete(UltraMesh* umesh) {
	ultrameshDestroy(umesh);
	free(umesh);
}

void ultrameshConstruct(UltraMesh* umesh) {
	vec_Construct(&umesh->layers, sizeof(UltraLayer));
	vec_Construct(&umesh->materials, sizeof(UltraMaterial));
	vec_Construct(&umesh->textures, sizeof(uint16_t));
}

void ultrameshDestroy(UltraMesh* umesh) {
	vec_Destroy(&umesh->textures);
	vec_Destroy(&umesh->materials);

 	for(uint32_t i=0, n=vec_GetSize(&umesh->layers); i<n; i++) {
		ultralayerDestroy((UltraLayer*)vec_GetElement(&umesh->layers, i));
	}
	vec_Destroy(&umesh->layers);
}

inline uint16_t ultrameshGetLayerCount(const UltraMesh* umesh) {
	return vec_GetSize(&umesh->layers);
}

inline uint16_t ultrameshGetMaterialCount(const UltraMesh* umesh) {
	return vec_GetSize(&umesh->materials);
}

inline uint16_t ultrameshGetTextureCount(const UltraMesh* umesh) {
	return vec_GetSize(&umesh->textures);
}

inline UltraLayer* ultrameshGetLayer(UltraMesh* umesh, uint16_t i) {
	return vec_GetElement(&umesh->layers, i);
}

inline uint16_t ultrameshGetTexture(const UltraMesh* umesh, uint16_t i) {
	return *(const uint16_t*)vec_GetConstElement(&umesh->textures, i);
}

inline const UltraLayer* ultrameshGetConstLayer(const UltraMesh* umesh, uint16_t i) {
	return vec_GetConstElement(&umesh->layers, i);
}

const UltraMaterial* ultrameshGetConstMaterial(const UltraMesh* umesh, uint16_t i) {
	return vec_GetConstElement(&umesh->materials, i);
}

/* inline const Image* ultrameshGetConstImage(const UltraMesh* umesh, uint16_t i) {
	return vec_GetConstElement(&umesh->textures, i);
} */

inline void ultrameshAddLayer(UltraMesh* umesh, const UltraLayer* layer) {
	vec_PushBack(&umesh->layers, layer);
}

inline void ultrameshAddMaterial(UltraMesh* umesh, const UltraMaterial* material) {
	vec_PushBack(&umesh->materials, material);
}

inline void ultrameshAddTexture(UltraMesh* umesh, uint16_t texid) {
	vec_PushBack(&umesh->textures, &texid);
}

static const char g_vts_begin[] = "static Vtx vts%d[] = {\n";
static const char g_vts_end[] = "};\n\n";
static const char g_gfx_begin[] = "static Gfx drawmesh%d_%d[] = {\n";
static const char g_gfx_end[] = "\tgsSPEndDisplayList()\n};\n\n";
static const char g_mat_begin[] = "static Material materials[] = {\n";
static const char g_mat_end[] = "};\n\n";
static const char g_pair_begin[] = "Pair tab[] = {\n";
static const char g_pair_end[] = "};\n\n";
static const char g_textab_begin[] = "static Texture textures[] = {\n";
static const char g_textab_content[] = "\t{%d, %d, &texture%d, %s},\n";
static const char g_textab_end[] = "};\n\n";
static const char g_tex_begin[] = "static %s texture%d[] = {\n";
static const char g_tex_end[] = "\n};\n\n";
#include "vec_pvoid.h"
#include "pairclipimage.h"
extern Vec(Pvoid) g_imagevec;

void ultrameshGenerateDLLayers(const UltraMesh* umesh, Vec(char)* charvec) {
	/* layers */
	for(uint16_t ilay=0, nlay=ultrameshGetLayerCount(umesh); ilay<nlay; ilay++) {
		const UltraLayer* ulayer = ultrameshGetConstLayer(umesh, ilay);

		/* verts */
		vecchar_sprintf(charvec, g_vts_begin, ilay);
		for(uint16_t iv=0, nv=ultralayerGetVertexCount(ulayer); iv<nv; iv++) {
			const UltraVertex* uvtx = ultralayerGetConstVertex(ulayer, iv);

 			vecchar_sprintf(charvec, "\t{\n\t\t{ {%d, %d, %d}, 0, {%d, %d}, {0x%x, 0x%x, 0x%x, 0xFF} }\n\t},\n",
				uvtx->x, uvtx->y, uvtx->z, uvtx->u, uvtx->v, uvtx->r&0xFF, uvtx->g&0xFF, uvtx->b&0xFF);
		}
		vecchar_sprintf(charvec, g_vts_end);

		/* faces */
		vecchar_sprintf(charvec, g_gfx_begin, ilay, 0);
		for(uint16_t iface=0, nface=ultralayerGetFaceCount(ulayer); iface<nface; iface++) {
			const UltraFace* uface = ultralayerGetConstFace(ulayer, iface);

			uint8_t n = ultrafaceGetVertexCount(uface);
			for(uint8_t i=0; i<n; i++)
				vecchar_sprintf(charvec, "\tgsSPVertex(&vts%d[%d], 1, %d),\n", ilay, ultrafaceGetIndex(uface, i), i);
			if(3 == n)
				vecchar_sprintf(charvec, "\tgsSP1Triangle(0, 1, 2, 0),\n");
			else if(4 == n)
				vecchar_sprintf(charvec, "\tgsSP2Triangles(0, 1, 2, 0, 0, 2, 3, 0),\n");
			else
				abort();
		}
		vecchar_sprintf(charvec, g_gfx_end);

	}
}

void ultrameshGenerateDLTextures(const UltraMesh* umesh, Vec(char)* charvec) {
	const uint16_t n=ultrameshGetTextureCount(umesh);
	if(0 == n)
		return;

	Vec(char) texattr;
	vecConstruct(char, &texattr);
	vecchar_sprintf(&texattr, g_textab_begin);
	for(uint16_t i=0; i<n; i++) {
		const PairClipImage* pair = *vecGetConstElement(Pvoid, &g_imagevec, i);
		const Image* image = pair->image;
		if(NULL == image) {
			vecchar_sprintf(&texattr, "\t{0, 0, NULL, 0},\n");
			continue;
		}

		uint16_t texid = ultrameshGetTexture(umesh, i);

		const char* fmt = NULL;
		if(IMAGE_FMT_RGBA32 == image->format) {
			vecchar_sprintf(charvec, g_tex_begin, "u32", texid, "\t");
			fmt = "TEXTURE_FMT_RGBA32";

			uint32_t k=0;
			for(uint32_t j=0; j<image->height; j++) {
				for(uint32_t i=0; i<image->width; i++) {
					if(0 == k%8)
						vecchar_sprintf(charvec, "\n\t");
					vecchar_sprintf(charvec, "0x%x, ", ((uint32_t*)image->pixels)[k++]);
				}
			}
		} else if(IMAGE_FMT_RGBA16 == image->format) {
			vecchar_sprintf(charvec, g_tex_begin, "u16", texid, "\t");
			fmt = "TEXTURE_FMT_RGBA16";

			if(imageGetPixelCount(image) > 2048)
				printf ( "Warning: texture %d won't fit in TMEM\n", texid );
			uint32_t k=0;
			for(uint32_t j=0; j<image->height; j++) {
				for(uint32_t i=0; i<image->width; i++) {
					if(0 == k%8)
						vecchar_sprintf(charvec, "\n\t");
					vecchar_sprintf(charvec, "0x%x, ", ((uint16_t*)image->pixels)[k++]);
				}
			}

		}

		vecchar_sprintf(&texattr, g_textab_content, image->width, image->height, texid, fmt);
		vecchar_sprintf(charvec, g_tex_end);
	}
	vecchar_sprintf(&texattr, g_textab_end);
	vecchar_sprintf(charvec, "%s", texattr.data);
	vecDestroy(char, &texattr);
}

void ultrameshGenerateDLMaterials(const UltraMesh* umesh, Vec(char)* charvec) {
	vecchar_sprintf(charvec, g_mat_begin);
	for(uint16_t i=0, n=ultrameshGetMaterialCount(umesh); i<n; i++) {
		const UltraMaterial* umaterial = ultrameshGetConstMaterial(umesh, i);
		const Vector3f* color = ultramaterialGetConstColor(umaterial);
		uint16_t texid = ultramaterialGetTexture(umaterial);
		
		vecchar_sprintf(charvec, "\t{ {%f, %f, %f}, ", color->x, color->y, color->z);
		if(ULTRAMATERIAL_TEXTURE_NONE == texid) {
			vecchar_sprintf(charvec, "NULL },\n");
		} else {
			vecchar_sprintf(charvec, "&textures[%d] },\n", texid);
		}

	}
	vecchar_sprintf(charvec, g_mat_end);
}

char* ultrameshGenerateDL(const UltraMesh* umesh) {
	Vec(char) charvec;
	vecConstruct(char, &charvec);

	ultrameshGenerateDLLayers(umesh, &charvec);
	ultrameshGenerateDLTextures(umesh, &charvec);
	ultrameshGenerateDLMaterials(umesh, &charvec);

	/* (Material, Gfx) pairs */
	vecchar_sprintf(&charvec, g_pair_begin);
	for(uint16_t ilay=0, nlay=ultrameshGetLayerCount(umesh); ilay<nlay; ilay++) {
		const UltraLayer* ulayer = ultrameshGetConstLayer(umesh, ilay);
		uint16_t matid = ultralayerGetMaterialId(ulayer);
		vecchar_sprintf(&charvec, "\t{&materials[%d], drawmesh%d_%d},\n", matid, ilay, 0);
	}
	vecchar_sprintf(&charvec, "\t{NULL, NULL}\n%s", g_pair_end);


	/* return copy */
	size_t vecsize = vecGetSize(char, &charvec);
	char* retdata = malloc(vecsize+1);
	if(NULL == retdata)
		return retdata;
	memcpy(retdata, charvec.data, vecsize+1); 
	vecDestroy(char, &charvec);

	return retdata;
}

